% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008-2010, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(drs_fol_pnf, [
		drs_fol/2,
		drs_pnf/2,
		drs_fol_pnf/3,
		fol_to_folpp/2
	]).

/** <module> Transform DRSs into First-order Formulas and Their Prenex Normal Forms

Technical Details:

drs_to_fol_to_prenex(+DRS, -FOL, -PRN) converts a discourse representation structure DRS to its   
equivalent first-order formula FOL and to its prenex normal form PNF.

Syntax of DRS: negation '-', disjunction 'v', conjunction ',', implication '=>', discourse referents 
are lists of existentially - in preconditions of implications universally - quantified variables.

Syntax of FOL and PRNF: negation '-', disjunction 'v', conjunction '&', implication '=>', universal 
quantification 'forall(X,Formula)', existential quantification 'exists(X,Formula)', where 'X' is a 
Prolog variable.

@author Norbert E. Fuchs
@author Kaarel Kaljurand
@author Tobias Kuhn
@version 2009-11-18
*/


:- op(400,  fy, -).   % negation
:- op(500, xfy, &).   % conjunction
:- op(600, xfy, v).   % disjunction
:- op(650, xfy, =>).  % implication 


%% drs_fol(+DRS:drs, -FOL:fol) is det.
%
% @param DRS is Attempto DRS
% @param FOL is FOL formula

drs_fol(drs([], []), '') :-
	!.

drs_fol(DRS, FOL) :-
	drs_to_fol(DRS, FOL),
	!.

drs_fol(_, '').


%% drs_pnf(+DRS:drs, -PNF:fol) is det.
%
% @param DRS is Attempto DRS
% @param PNF is FOL formula in PNF

drs_pnf(drs([], []), '') :-
	!.

drs_pnf(DRS, PNF) :-
	drs_to_fol(DRS, FOL),
	prenex_normal_form(FOL, PNF),
	!.

drs_pnf(_, '').


%% drs_fol_pnf(+DRS:drs, -FOL:fol, -PNF:fol) is det.
%
% Converts DRS to first-order formula FOL and to prenex normal form PNF.
%
% @param DRS is Attempto DRS
% @param FOL is FOL formula
% @param PNF is FOL formula in PNF

drs_fol_pnf(drs([], []), '', '') :- !.

drs_fol_pnf(DRS, FOL, PNF) :-
	drs_to_fol(DRS, FOL),
	prenex_normal_form(FOL, PNF),
	!.

drs_fol_pnf(_, '', '').


%% fol_to_folpp(+Fol:term, -FolPp:atom) is det.
%
% Prints the FOL formula (which can be in PNF) into an atom.
% The printing is done with write/1 which respects the
% operator definitions, i.e. binary operators are printed in infix form.
%
% @param Fol is a FOL formula (possibly in PNF)
% @param FolPp is the FOL formula pretty-printed

fol_to_folpp(Fol, FolPp) :-
	copy_term(Fol, FolCopy),
	numbervars(FolCopy, 0, _),
	with_output_to(atom(FolPp), write_term(FolCopy, [numbervars(true), quoted(true), module(drs_fol_pnf)])).


%% drs_to_fol(+DRS, -FOL)
%
% convert DRS to first-order formula FOL
%
% based on drs2fol.pl (P. Blackburn & J. Bos, From DRSs to First-Order Logic, 1999) 

drs_to_fol(drs([],[Cond]), Formula) :-
	!,
	cond_to_fol(Cond, Formula).

drs_to_fol(drs([],[Cond1,Cond2|Conds]), Formula1 & Formula2) :-
	!,
	cond_to_fol(Cond1, Formula1),
	drs_to_fol(drs([],[Cond2|Conds]), Formula2).

drs_to_fol(drs([X|Referents],Conds), exists(X,Formula)) :-
	!,
	drs_to_fol(drs(Referents,Conds), Formula).

cond_to_fol(-Drs, -Formula) :-
	!,
	drs_to_fol(Drs, Formula).

cond_to_fol(Conds, Formula) :-
	is_list(Conds),
	!,
	drs_to_fol(drs([],Conds), Formula).

cond_to_fol(Drs1 v Drs2, Formula1 v Formula2) :-
	!,
	drs_to_fol(Drs1,Formula1),
	drs_to_fol(Drs2,Formula2).

cond_to_fol(drs([],Conds) => Drs2, Formula1 => Formula2) :-
	!,
	drs_to_fol(drs([],Conds),Formula1),
	drs_to_fol(Drs2,Formula2).

cond_to_fol(drs([X|Referents],Conds) => Drs2, forall(X,Formula)) :-
	!,
	cond_to_fol(drs(Referents,Conds) => Drs2,Formula).

cond_to_fol(BasicCondition - Index, BasicCondition - Index). 
   

%% prenex_normal_form(+FOL, -PNF)
%
% convert first-order formula FOL into its prenex normal form PNF
%
% based on transform/2 (P. A. Flach, Simply Logical: Intelligent Reasoning by Example, John Wiley & Sons, 1994)

prenex_normal_form(FOL, PNF):-
	rewrite_implications(FOL, F1),
	negations_inside(F1, F2),
	prenex_form(F2, PNF).


%% rewrite implications(+In, -Out)
% 

rewrite_implications(A, A):-						
	literal(A).

rewrite_implications(A => B, -C v D):-			
	rewrite_implications(A, C),
	rewrite_implications(B, D).

rewrite_implications(A & B, C & D):-			
	rewrite_implications(A, C),					
	rewrite_implications(B, D).

rewrite_implications(A v B, C v D):-
	rewrite_implications(A, C),
	rewrite_implications(B, D).

rewrite_implications(-A, -C):-
	rewrite_implications(A, C).

rewrite_implications(forall(X, B), forall(X, D)):-
	rewrite_implications(B, D).

rewrite_implications(exists(X, B), exists(X, D)):-
	rewrite_implications(B, D).


%% negations_inside(+In, -Out)
%
% move negations inside

negations_inside(A, A):-							
	literal(A).

negations_inside(-(A & B), C v D):-				
	negations_inside(-A, C),
	negations_inside(-B, D).

negations_inside(-(A v B), C & D):-				
	negations_inside(-A, C),
	negations_inside(-B, D).

negations_inside(-(-A), B):-						
	negations_inside(A, B).

negations_inside(-exists(X, A), forall(X, B)):-	
	negations_inside(-A, B).

negations_inside(-forall(X, A), exists(X, B)):-
	negations_inside(-A, B).

negations_inside(A & B, C & D):-				
	negations_inside(A, C),						
	negations_inside(B, D).

negations_inside(A v B, C v  D):-
	negations_inside(A, C),
	negations_inside(B, D).

negations_inside(exists(X, A), exists(X, B)):-
	negations_inside(A, B).

negations_inside(forall(X, A), forall(X, B)):-
	negations_inside(A, B).


%% prenex_form(+In, -Out)
%
% generate prenex normal form

prenex_form(F, PNF):-
	pnf(F, PNF, B, B).


pnf(A, V, V, A):-									
	literal(A).

pnf(forall(X, F), forall(X, Quants), V, Body):-
	pnf(F, Quants, V, Body).

pnf(exists(X, F), exists(X, Quants), V, Body):-
	pnf(F, Quants, V, Body).

pnf(A & B, Quants, V, BodyA & BodyB):-
	pnf(A, Quants, QB, BodyA),
	pnf(B, QB, V, BodyB).

pnf(A v B, Quants, V, BodyA v BodyB):-
	pnf(A, Quants, QB, BodyA),
	pnf(B, QB, V, BodyB).


%% disjunction_of_literals(+Term)
%

disjunction_of_literals(A) :-
	literal(A).

disjunction_of_literals(A v B):-
	disjunction_of_literals(A),
	disjunction_of_literals(B).


%% literal(+Term)
%

literal(_A - _Index).

literal(-(_A - _Index)).
