% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Kaarel Kaljurand <kaljurand@gmail.com>.
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(implication_to_swrl, [
		implication_to_swrl/3
	]).

:- use_module(drs_to_owlswrl_core, [
		is_toplevel/3,
		get_namedclass/2,
		get_namedobjectproperty/2,
		get_nameddataproperty/2,
		condlist_and/4,
		is_object_with_generalized_quantifier/1
	]).


/** <module> DRS implication to SWRL 'Implies'-rule translator

Translate an Attempto DRS implication into Semantic Web Rule Language (SWRL) 'Implies'-rule.

@author Kaarel Kaljurand
@version 2009-05-17
*/


%% implication_to_swrl(+Condition:term, +RefList:list, -Implies:term) is semidet.
%
% @param Condition is a DRS implication condition
% @param RefList is a list of DRS toplevel referents
% @param Implies is a SWRL 'Implies'-rule
%
implication_to_swrl(
	'=>'(CondList1, CondList2),
	RefList,
	'Implies'('Antecedent'(AtomList1), 'Consequent'(AtomList2))
) :-
	condlist_atomlist(CondList1, RefList, AtomList1),
	condlist_atomlist(CondList2, RefList, AtomList2).


%% condlist_atomlist(+ConditionList:list, +RefList:list, -AtomList:list) is det.
%
% @param ConditionList is a list of DRS conditions
% @param RefList is a list of DRS toplevel referents
% @param AtomList is list of SWRL atoms
%
condlist_atomlist([], _, []).

condlist_atomlist([Condition | ConditionList], RefList, [Atom | AtomList]) :-
	condition_atom(Condition, RefList, Atom),
	condlist_atomlist(ConditionList, RefList, AtomList).


/*
% BUG: experimental
% BUG: Add a restriction that all variables used in predicate-conditions
% must be declared in object-conditions.


condlist_atomlist([], _RefList, []).

condlist_atomlist([Condition | CondList], RefList, [Atom | AtomList]) :-
	(
		Condition = object(D, Value, countable,  _, _, _)-_
	->
		get_namedclass(Value, NamedClass),
		select(Condition1, CondList, CondList1),
		condlist_class(D, Condition1, CondList1, RefList, AndClass, CondList2),
		Atom = 'description'(and(NamedClass, AndClass), 'I-variable'(D))
	;
		condition_atom(Condition, RefList, Atom),
		CondList2 = CondList
	),
	condlist_atomlist(CondList2, RefList, AtomList).
*/



%% condition_atom(+Condition:term, -Atom:term) is det.
%
% @param Condition is a DRS condition
% @param RefList is a list of DRS toplevel referents
% @param Atom is a SWRL atom
%

% We allow no generalized quantifiers
condition_atom(Condition, _, _) :-
	is_object_with_generalized_quantifier(Condition),
	!,
	fail.

condition_atom(object(X, Value, _,  _, _, _)-_, _, 'description'(NamedClass, 'I-variable'(X))) :-
	get_namedclass(Value, NamedClass).

condition_atom('-'([predicate(_, be, X1, X2)-_]), RefList, 'differentFrom'(OutX1, OutX2)) :-
	get_argument(X1, RefList, object, OutX1),
	get_argument(X2, RefList, object, OutX2).

condition_atom(predicate(_, be, X1, X2)-_, RefList, 'sameAs'(OutX1, OutX2)) :-
	get_argument(X1, RefList, object, OutX1),
	get_argument(X2, RefList, object, OutX2).

condition_atom(predicate(_, Value, X1, X2)-_, RefList, 'description'(DataProperty, OutX1, OutX2)) :-
	Value \= be,
	get_nameddataproperty(Value, DataProperty),
	get_argument(X1, RefList, Type1, OutX1),
	get_argument(X2, RefList, Type2, OutX2),
	(
		Type1 = data
	;
		Type2 = data
	).

condition_atom(predicate(_, Value, X1, X2)-_, RefList, 'description'(ObjectProperty, OutX1, OutX2)) :-
	Value \= be,
	get_namedobjectproperty(Value, ObjectProperty),
	get_argument(X1, RefList, object, OutX1),
	get_argument(X2, RefList, object, OutX2).

condition_atom(formula(X1, Op, X2)-_, RefList, Atom) :-
	get_argument(X1, RefList, _, OutX1),
	get_argument(X2, RefList, _, OutX2),
	op_builtin(Op, [OutX1, OutX2], Atom).

% Sublist
% E.g. If a man owns a car then the man likes the car and likes exactly 3 cats.
% E.g. If John owns a car then Mary likes a car and likes at most 3 cats.
condition_atom([C | Cs], RefList, 'description'(SubListClass, OutX)) :-
	condlist_and(X, [C | Cs], RefList, SubListClass),
	get_argument(X, RefList, object, OutX).

% Note that X can point to an Individual (i.e. not only a variable),
% see e.g. "If John does not like Mary then a man does not like a woman."
condition_atom('-'(Not), RefList, 'description'('ObjectComplementOf'(NotClass), OutX)) :-
	condlist_and(X, Not, RefList, NotClass),
	get_argument(X, RefList, object, OutX).

% Note that X can point to an Individual.
condition_atom('v'(Or1, Or2), RefList, 'description'('ObjectUnionOf'([Or1Class, Or2Class]), OutX)) :-
	condlist_and(X, Or1, RefList, Or1Class),
	condlist_and(X, Or2, RefList, Or2Class),
	get_argument(X, RefList, object, OutX).


%% get_argument(+X:term, +RefList:list, -Type:atom, -Argument:term) is det.
%
% @param X is a discourse referent
% @param RefList is a list of DRS toplevel referents
% @param Type is the type of the argument, i.e. one of {data, object, _}
% @param Argument is a SWRL variable, OWL individual, or OWL data value
%
get_argument(expr(X, Y, Z), _RefList, data, expr(X, Y, Z)) :- !.

get_argument(int(X), _RefList, data, X) :- !.

get_argument(real(X), _RefList, data, X) :- !.

get_argument(string(X), _RefList, data, X) :- !.

get_argument(X, RefList, object, Individual) :-
	is_toplevel(X, RefList, 'ObjectOneOf'([Individual])),
	!.

get_argument(X, _RefList, object, 'I-variable'(X)).


%% op_builtin(+Relation:atom, +ArgumentList:list, -Builtin:term) is det.
%
% @param Relation is one of {'>', '<', '='}
% @param ArgumentList is a list of relation arguments
% @param Builtin is a term containing the relation symbol and the list of arguments
%
op_builtin('>', [X1, X2], builtIn(greaterThan, X1, X2)) :- !.

op_builtin('<', [X1, X2], builtIn(lessThan, X1, X2)) :- !.

op_builtin('=', [X1, expr('+', X2, X3)], builtIn(add, X1, X2, X3)) :- !.
op_builtin('=', [X1, expr('-', X2, X3)], builtIn(subtract, X1, X2, X3)) :- !.
op_builtin('=', [X1, expr('*', X2, X3)], builtIn(multiply, X1, X2, X3)) :- !.
op_builtin('=', [X1, expr('/', X2, X3)], builtIn(divide, X1, X2, X3)) :- !.

op_builtin('=', [expr('+', X2, X3), X1], builtIn(add, X1, X2, X3)) :- !.
op_builtin('=', [expr('-', X2, X3), X1], builtIn(subtract, X1, X2, X3)) :- !.
op_builtin('=', [expr('*', X2, X3), X1], builtIn(multiply, X1, X2, X3)) :- !.
op_builtin('=', [expr('/', X2, X3), X1], builtIn(divide, X1, X2, X3)) :- !.

op_builtin('=', [X1, X2], builtIn(equal, X1, X2)).
